@@ -1,17 +1,19 @@
-from typing import Set, Optional, Dict, Protocol
+from typing import Set, Optional, Dict, Protocol, Callable, TypeVar, Any, ClassVar, cast
 from dataclasses import dataclass
 from abc import ABC, abstractmethod
+from functools import wraps
 
 from starlette.types import ASGIApp
 from starlette.requests import Request
 from starlette.responses import Response
 from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
 
-from .errors import Unauthorised
+from .errors import Unauthorised, Forbidden
 
 @dataclass
 class Principal:
 	id: str
+	privileges: Set[str]
 
 class AuthBackend(Protocol):
 	async def resolve_principal(self, req: Request) -> Optional[Principal]:
@@ -31,8 +33,7 @@ class AuthMiddleware(BaseHTTPMiddleware):
 		return await call_next(req)
 
 class HeaderAuthBackend(ABC):
-	def __init__(self, header: str) -> None:
-		self.header = header
+	header: ClassVar[str]
 
 	async def resolve_principal(self, req: Request) -> Optional[Principal]:
 		token = req.headers.get(self.header)
@@ -44,3 +45,18 @@ class HeaderAuthBackend(ABC):
 	@abstractmethod
 	async def principal_from_header(self, value: str) -> Optional[Principal]:
 		...
+
+TFunc = TypeVar('TFunc', bound = Callable[..., Any])
+
+def privileged(privilege: str) -> Callable[[ TFunc ], TFunc]:
+	def _decorator(func: TFunc) -> TFunc:
+		@wraps(func)
+		async def _impl(*args: Any, **kwargs: Any) -> Any:
+			principal = cast(Principal, kwargs['principal'])
+			if privilege not in principal.privileges:
+				raise Forbidden('insufficient privileges')
+
+			return await func(*args, **kwargs)
+
+		return cast(TFunc, _impl)
+	return _decorator
